module System.Win32.Service where

import System.Win32.Types

import Foreign
import Foreign.C

#include <windows.h>

type SC_HANDLE = Ptr ()
type ServiceHandle = SC_HANDLE

type ServiceAccessRights = DWORD
#{enum ServiceAccessRights,
  , sC_MANAGER_ALL_ACCESS         = SC_MANAGER_ALL_ACCESS
  , sC_MANAGER_CREATE_SERVICE     = SC_MANAGER_CREATE_SERVICE
  , sC_MANAGER_CONNECT            = SC_MANAGER_CONNECT
  , sC_MANAGER_ENUMERATE_SERVICE  = SC_MANAGER_ENUMERATE_SERVICE
  , sC_MANAGER_LOCK               = SC_MANAGER_LOCK
  , sERVICE_ALL_ACCESS            = SERVICE_ALL_ACCESS
  , sERVICE_CHANGE_CONFIG         = SERVICE_CHANGE_CONFIG
  , sERVICE_ENUMERATE_DEPENDENTS  = SERVICE_ENUMERATE_DEPENDENTS
  , sERVICE_INTERROGATE           = SERVICE_INTERROGATE
  , sERVICE_PAUSE_CONTINUE        = SERVICE_PAUSE_CONTINUE
  , sERVICE_QUERY_CONFIG          = SERVICE_QUERY_CONFIG
  , sERVICE_QUERY_STATUS          = SERVICE_QUERY_STATUS
  , sERVICE_START                 = SERVICE_START
  , sERVICE_STOP                  = SERVICE_STOP
  , sERVICE_USER_DEFINED_CONTROL  = SERVICE_USER_DEFINED_CONTROL
  , aCCESS_SYSTEM_SECURITY        = ACCESS_SYSTEM_SECURITY
  , dELETE                        = DELETE
  , rEAD_CONTROL                  = READ_CONTROL
  , wRITE_DAC                     = WRITE_DAC
  , wRITE_OWNER                   = WRITE_OWNER
  }

type ServiceType = DWORD
#{enum ServiceType,
  , sERVICE_FILE_SYSTEM_DRIVER    = SERVICE_FILE_SYSTEM_DRIVER
  , sERVICE_KERNEL_DRIVER         = SERVICE_KERNEL_DRIVER
  , sERVICE_WIN32_OWN_PROCESS     = SERVICE_WIN32_OWN_PROCESS
  , sERVICE_WIN32_SHARE_PROCESS   = SERVICE_WIN32_SHARE_PROCESS
  }

type ServiceStartType = DWORD
#{enum ServiceStartType,
  , sERVICE_AUTO_START      = SERVICE_AUTO_START
  , sERVICE_BOOT_START      = SERVICE_BOOT_START
  , sERVICE_DEMAND_START    = SERVICE_DEMAND_START
  , sERVICE_DISABLED        = SERVICE_DISABLED
  , sERVICE_SYSTEM_START    = SERVICE_SYSTEM_START
  }

type ServiceErrorControl = DWORD
#{enum ServiceErrorControl,
  , sERVICE_ERROR_CRITICAL  = SERVICE_ERROR_CRITICAL
  , sERVICE_ERROR_IGNORE    = SERVICE_ERROR_IGNORE
  , sERVICE_ERROR_NORMAL    = SERVICE_ERROR_NORMAL
  , sERVICE_ERROR_SEVERE    = SERVICE_ERROR_SEVERE
  }

type ServiceControl = DWORD
#{enum ServiceControl,
  , sERVICE_CONTROL_CONTINUE        = SERVICE_CONTROL_CONTINUE
  , sERVICE_CONTROL_INTERROGATE     = SERVICE_CONTROL_INTERROGATE
  , sERVICE_CONTROL_NETBINDADD      = SERVICE_CONTROL_NETBINDADD
  , sERVICE_CONTROL_NETBINDDISABLE  = SERVICE_CONTROL_NETBINDDISABLE
  , sERVICE_CONTROL_NETBINDENABLE   = SERVICE_CONTROL_NETBINDENABLE
  , sERVICE_CONTROL_NETBINDREMOVE   = SERVICE_CONTROL_NETBINDREMOVE
  , sERVICE_CONTROL_PARAMCHANGE     = SERVICE_CONTROL_PARAMCHANGE
  , sERVICE_CONTROL_PAUSE           = SERVICE_CONTROL_PAUSE
  , sERVICE_CONTROL_STOP            = SERVICE_CONTROL_STOP
  }

type ServiceStatus = ()

type TagId = DWORD

type ServiceMainClosure = DWORD -> [String] -> IO ()

type C_ServiceMainClosure = DWORD -> Ptr (Ptr CChar) -> IO ()

changeServiceConfig
  :: SC_HANDLE
  -> ServiceType
  -> ServiceStartType
  -> ServiceErrorControl
  -> Maybe String
  -> Maybe String
  -> Maybe String
  -> Maybe String
  -> Maybe String
  -> Maybe String
  -> IO TagId
changeServiceConfig
  hService serviceType startType errorControl binaryPathName loadOrderGroup
  dependencies startName password displayName =
    maybeWith withTString binaryPathName $ \ c_binaryPathName ->
    maybeWith withTString loadOrderGroup $ \ c_loadOrderGroup ->
    maybeWith withTString dependencies $ \ c_dependencies ->
    maybeWith withTString startName $ \ c_startName ->
    maybeWith withTString password $ \ c_password ->
    maybeWith withTString displayName $ \ c_displayName ->
    alloca $ \p_tag -> do
      failIfFalse_ "ChangeServiceConfig" $ c_ChangeServiceConfig
        hService serviceType startType errorControl c_binaryPathName
        c_loadOrderGroup p_tag c_dependencies c_startName c_password
        c_displayName
      peek p_tag
foreign import stdcall unsafe "windows.h ChangeServiceConfigW"
  c_ChangeServiceConfig
    :: SC_HANDLE
    -> ServiceType
    -> ServiceStartType
    -> ServiceErrorControl
    -> LPCTSTR
    -> LPCTSTR
    -> Ptr DWORD
    -> LPCTSTR
    -> LPCTSTR
    -> LPCTSTR
    -> LPCTSTR
    -> IO Bool

{-- Omitted: ChangeServiceConfig2 --}

closeServiceHandle :: SC_HANDLE -> IO ()
closeServiceHandle hService =
  failIfFalse_ "CloseServiceHandle" $ c_CloseServiceHandle hService
foreign import stdcall unsafe "windows.h CloseServiceHandle"
  c_CloseServiceHandle :: SC_HANDLE -> IO Bool

controlService :: SC_HANDLE -> ServiceControl -> Ptr ServiceStatus -> IO ()
controlService hService control p_serviceStatus =
  failIfFalse_ "ControlService" $
    c_ControlService hService control p_serviceStatus
foreign import stdcall unsafe "windows.h ControlService"
  c_ControlService
    :: SC_HANDLE -> ServiceControl -> Ptr ServiceStatus -> IO Bool

createService
  :: SC_HANDLE
  -> String
  -> String
  -> ServiceAccessRights
  -> ServiceType
  -> ServiceStartType
  -> ServiceErrorControl
  -> String
  -> Maybe String
  -> Ptr TagId
  -> Maybe String
  -> Maybe String
  -> Maybe String
  -> IO SC_HANDLE
createService
  hSCManager serviceName displayName desiredAccess serviceType startType
  errorControl binaryPathName loadOrderGroup p_tagId
  dependencies serviceStartName password =
    withTString serviceName $ \ c_serviceName ->
    withTString displayName $ \ c_displayName ->
    withTString binaryPathName $ \ c_binaryPathName ->
    maybeWith withTString loadOrderGroup $ \ c_loadOrderGroup ->
    maybeWith withTString dependencies $ \ c_dependencies ->
    maybeWith withTString serviceStartName $ \ c_serviceStartName ->
    maybeWith withTString password $ \ c_password -> do
      failIfNull "CreateService" $ c_CreateService
        hSCManager c_serviceName c_displayName desiredAccess serviceType
        startType errorControl c_binaryPathName c_loadOrderGroup p_tagId
        c_dependencies c_serviceStartName c_password
foreign import stdcall unsafe "windows.h CreateServiceW"
  c_CreateService
    :: SC_HANDLE
    -> LPCTSTR
    -> LPCTSTR
    -> DWORD
    -> DWORD
    -> DWORD
    -> DWORD
    -> LPCTSTR
    -> LPCTSTR
    -> Ptr TagId
    -> LPCTSTR
    -> LPCTSTR
    -> LPCTSTR
    -> IO SC_HANDLE

deleteService :: SC_HANDLE -> IO ()
deleteService hService = failIfFalse_ "DeleteService" $ c_DeleteService hService
foreign import stdcall unsafe "windows.h DeleteService"
  c_DeleteService :: SC_HANDLE -> IO Bool

{-- Omitted: EnumDependentServices --}

{-- Omitted: EnumServicesStatus --}

{-- Omitted: EnumServicesStatusEx --}

{-- Omitted: GetServiceDisplayName --}

{-- Omitted: GetServiceKeyName --}

{-- Omitted: Handler --}

{-- Omitted: HandlerEx --}

{-- Omitted: LockServiceDatabase --}

{-- Omitted: NotifyBootConfigStatus --}

openSCManager
  :: Maybe String
  -> Maybe String
  -> ServiceAccessRights
  -> IO SC_HANDLE
openSCManager machineName databaseName desiredAccess =
  maybeWith withTString machineName $ \ c_machineName ->
  maybeWith withTString databaseName $ \ c_databaseName ->
    failIfNull "OpenSCManager" $
      c_OpenSCManager c_machineName c_databaseName desiredAccess
foreign import stdcall unsafe "windows.h OpenSCManagerW"
  c_OpenSCManager :: LPCTSTR -> LPCTSTR -> ServiceAccessRights -> IO SC_HANDLE

openService :: SC_HANDLE -> String -> ServiceAccessRights -> IO SC_HANDLE
openService hSCManager serviceName desiredAccess =
  withTString serviceName $ \ c_serviceName ->
    failIfNull "OpenService" $
      c_OpenService hSCManager c_serviceName desiredAccess
foreign import stdcall unsafe "windows.h OpenService"
  c_OpenService :: SC_HANDLE -> LPCTSTR -> ServiceAccessRights -> IO SC_HANDLE

{-- Omitted: QueryServiceConfig --}

{-- Omitted: QueryServiceConfig2 --}

{-- Omitted: QueryServiceLockStatus --}

{-- Omitted: QueryServiceStatus --}

{-- Omitted: QueryServiceStatusEx --}

{-- Omitted: RegisterServiceCtrlHandler --}

{-- Omitted: RegisterServiceCtrlHandlerEx --}

{-- Omitted: ServiceMain --}

{-- Omitted: SetServiceBits --}

{-- Omitted: SetServiceStatus --}

{-- Omitted: StartService --}

{-- Omitted: StartServiceCtrlDispatcher --}
foreign import ccall "wrapper"
  mkServiceMainClosure
    :: C_ServiceMainClosure
    -> IO (FunPtr C_ServiceMainClosure)

{--
startServiceCtrlDispatcher :: [ServiceTableEntry] -> IO ()
startServiceCtrlDispatcher sericeTableEntries = do
  failIfFalse_ "StartServiceCtrlDispatcher" $
    c_StartServiceCtrlDispatcher sericeTableEntries
foreign import stdcall c_StartServiceCtrlDispatcher unsafe
  "windows.h StartServiceCtrlDispatcherW"
    :: [ServiceTableEntry] -> IO Bool
--}

{-- Omitted: UnlockServiceDatabase --}

